import bpy
from pathlib import Path

from typing import Iterable

from ..definitions import MODIFIER_NAME, ROOT_PACKAGE
from bpy_extras.id_map_utils import get_all_referenced_ids, get_id_reference_map # type: ignore
from _bpy import types as bpy_types # type: ignore

def exclude_collection(context, collection_name, exclude = True,):
    def traverse_tree(t):
        yield t
        for child in t.children:
            yield from traverse_tree(child)

    coll = context.view_layer.layer_collection

    for c in traverse_tree(coll):
        if c.name == collection_name:
            c.exclude = exclude

def get_root_identifier(bl_rna):
    identifiers = []
    while bl_rna is not None:
        identifiers.append(bl_rna.identifier)
        bl_rna = bl_rna.base
    return identifiers[-2] # last element will always be 'ID', we want direct sub-class of ID, hence -2

def get_blenddata_collection(data_block):
    id_blenddata_map = {
        "Action" : bpy.data.actions,
        "Armature" : bpy.data.armatures,
        "Brush" : bpy.data.brushes,
        "Camera" : bpy.data.cameras,
        "Collection" : bpy.data.collections,
        "Curve" : bpy.data.curves,
        "VectorFont" : bpy.data.fonts,
        "GreasePencil" : bpy.data.grease_pencils,
        "Curves" : bpy.data.hair_curves,
        "Image" : bpy.data.images,
        "Lattice" : bpy.data.lattices,
        "LightProbe" : bpy.data.lightprobes,
        "Light" : bpy.data.lights,
        "FreestyleLineStyle" : bpy.data.linestyles,
        "Mask" : bpy.data.masks,
        "Material" : bpy.data.materials,
        "Mesh" : bpy.data.meshes,
        "MetaBall" : bpy.data.metaballs,
        "MovieClip" : bpy.data.movieclips,
        "NodeTree" : bpy.data.node_groups,
        "Object" : bpy.data.objects,
        "Palette" : bpy.data.palettes,
        "ParticleSettings" : bpy.data.particles,
        "PointCloud" : bpy.data.pointclouds,
        "Scene" : bpy.data.scenes,
        "Screen" : bpy.data.screens,
        "Key" : bpy.data.shape_keys,
        "Speaker" : bpy.data.speakers,
        "Text" : bpy.data.texts,
        "Texture" : bpy.data.textures,
        "Volume" : bpy.data.volumes,
        "WindowManager" : bpy.data.window_managers,
        "WorkSpace" : bpy.data.workspaces,
        "World" : bpy.data.worlds,
    }

    identifier = get_root_identifier(data_block.bl_rna)
    return id_blenddata_map[identifier]

def get_original_blenddata(data_block):
    blenddata_collection = get_blenddata_collection(data_block) 
    name = data_block.name
    if '.' in name:
        # get original name without .000 : node_group.004 -> node_group
        og_name,suffix = name.rsplit('.',1)
        # check if node_group with og_name exist
        og_data_block = None
        if og_name in blenddata_collection:
            og_data_block = blenddata_collection[og_name]
        return og_data_block
    return None

class Procedural_Signs_Append_Mixin:
    bl_options = {"REGISTER","UNDO","INTERNAL"}

    @classmethod
    def poll(cls, context):
        return True
    
    def ensure_and_get_collection(self,collection_name:str):
        coll = bpy.data.collections.get(collection_name)
        if not coll:
            coll = bpy.data.collections.new(collection_name)
            bpy.context.scene.collection.children.link(coll)
        return coll

    def get_duplicate_node_group(self,name):
        if '.' in name:
            # get original name without .000 : node_group.004 -> node_group
            og_name,suffix = name.rsplit('.',1)
            # check if node_group with og_name exist
            og_node_group = None
            if og_name in bpy.data.node_groups:
                og_node_group = bpy.data.node_groups[og_name]
            return og_node_group
        return None

    def import_and_link_asset_collection(self,blend_file:Path,collection_name:str,parent_collection:bpy.types.Collection=None):
        with bpy.data.libraries.load(str(blend_file), link=False) as (data_from, data_to):
            data_to.collections = [coll for coll in data_from.collections if coll == collection_name]
        
        if parent_collection:
            for collection in data_to.collections:
                parent_collection.children.link(collection)

    def import_and_link_objects(self,blend_file:Path,object_names:Iterable,parent_collection:bpy.types.Collection=None):
        with bpy.data.libraries.load(str(blend_file), link=False) as (data_from, data_to):
            data_to.objects = [obj for obj in data_from.objects if obj in object_names]
        
        if parent_collection:
            for obj in data_to.objects:
                parent_collection.objects.link(obj)

    def get_or_import_node_group(self,blend_file:Path,node_group_name:str):
        # check if node-group exist 
        node_group = bpy.data.node_groups.get(node_group_name)
        is_exists = node_group is not None
        if not node_group: # if not exist
        # load gn node group
            with bpy.data.libraries.load(str(blend_file), link=False) as (data_from, data_to):
                data_to.node_groups = [ng for ng in data_from.node_groups if ng == node_group_name]

            node_group = bpy.data.node_groups.get(node_group_name)
            referenced_ids = get_all_referenced_ids(node_group, get_id_reference_map())

            orig_node_group = get_original_blenddata(node_group)
            if orig_node_group:
                bpy.data.node_groups.remove(node_group)
                node_group.user_remap(orig_node_group)
                node_group = orig_node_group

            for data_block in referenced_ids:
                orig_data_block = get_original_blenddata(data_block)
                if orig_data_block:
                    blenddata_collection = get_blenddata_collection(data_block)
                    data_block.user_remap(orig_data_block)
                    blenddata_collection.remove(data_block)
            # TODO: cleanup of dependent data of deleted nodegroup

        return node_group, is_exists
    
    def get_path_to_sign_blendfile(self,context):
        props = context.scene.procedural_signs
        asset_dir = context.preferences.addons[ROOT_PACKAGE].preferences.filepath
        signs_path = Path(asset_dir)/'Signs'/props.sign
        path_to_blend = next(signs_path.glob("*.blend"))
        return path_to_blend
    
    def setup_objects_with_geometry_nodegroup(self, context, objects:Iterable[bpy.types.Object], suffix="", link_to_collection:bool=True):
        props = context.scene.procedural_signs
        path_to_blend = self.get_path_to_sign_blendfile(context)
        coll = self.ensure_and_get_collection(f'Signs - {props.sign}')
        if link_to_collection:
            for obj in objects:
                coll.objects.link(obj)

        node_group, is_nodegroup_exists = self.get_or_import_node_group(path_to_blend,f'{props.sign}{suffix}')
        asset_coll = bpy.data.collections.get(f'{props.sign}_ASSETS')
        # asset_coll being None means, user may have renamed the group or deleted the collection.
        if not is_nodegroup_exists or asset_coll is None:
            referenced_ids = get_all_referenced_ids(node_group, get_id_reference_map())
            for id_ in referenced_ids:
                if isinstance(id_,(bpy.types.Object,bpy_types.Object)):
                    asset_coll = bpy.data.collections.get(f'{props.sign}_ASSETS')
                    if not asset_coll:
                        asset_coll = bpy.data.collections.new(f'{props.sign}_ASSETS')
                        coll.children.link(asset_coll)
                    if id_.name not in asset_coll.objects:
                        asset_coll.objects.link(id_)
            if asset_coll:
                exclude_collection(context,asset_coll.name,exclude=True)
        for obj in objects:
            modifier  = obj.modifiers.new(MODIFIER_NAME,type='NODES')
            modifier.node_group = node_group

    def execute(self, context):
        ...