from pathlib import Path
import typing
import bpy
from bpy.types import Context, Operator

from ..definitions import MODIFIER_NAME, OTHER_ASSET_COLLECTION, OTHER_MAIN_COLLECTION, PRESET_ASSET_COLLECTION, PRESET_MAIN_COLLECTION, ROOT_PACKAGE
from ..properties import _sign_metadata_map, sign_type
from .mixins import Procedural_Signs_Append_Mixin, exclude_collection

# https://github.com/dfelinto/blender/blob/master/release/scripts/modules/bpy_types.py


class Procedural_Signs_OT_append(Operator):
    bl_idname = "procedural_signs.append"
    bl_label = "Add Sign"
    bl_description = "Add Procedural Sign to the scene"
    bl_options = {"REGISTER","UNDO"}

    @classmethod
    def poll(cls, context) -> bool:
        if context.mode != 'OBJECT':
            cls.poll_message_set('You should be in Object mode!')
            return False
        props = context.scene.procedural_signs
        selected_sign_type = _sign_metadata_map[f'Signs_{props.sign}'].sign_type
        if selected_sign_type == sign_type.IMAGE.value \
            and props.image_usage_option == 'EXISTING' \
            and props.selected_image is None:
            cls.poll_message_set('Select Image!')
            return False
        if selected_sign_type == sign_type.OBJECT.value \
            and props.object_usage_option == 'SELECTED' \
            and (context.active_object is None or context.active_object.type != 'CURVE'):
            cls.poll_message_set('Select a Curve Object')
            return False
        return True

    def execute(self, context):
        props = context.scene.procedural_signs
        selected_sign_type = _sign_metadata_map[f'Signs_{props.sign}'].sign_type
        if selected_sign_type == sign_type.OBJECT.value:
            if props.object_usage_option == 'NEW' and props.import_option == 'SVG':
                # open filw browser to select svg file and import it
                # merge curves into one curve - optional
                # assign geometry node modifier with curve node group
                bpy.ops.procedural_signs.append_svg('INVOKE_DEFAULT')
            if props.object_usage_option == 'NEW' and props.import_option == 'DRAW':
                # create a empty curve object
                # assign geometry node modifier with curve node group
                # go to edit mode
                # go to top view
                # start draw_curve operator
                bpy.ops.procedural_signs.append_new_draw('INVOKE_DEFAULT')

            if props.object_usage_option == 'NEW' and props.import_option == 'TEXT':
                # create empty curve object
                # assign geometry node modifier with text node group
                bpy.ops.procedural_signs.append_new_text('INVOKE_DEFAULT')
            
            if props.object_usage_option == 'SELECTED':
                bpy.ops.procedural_signs.append_selected('INVOKE_DEFAULT') # invoke is not necessary 

        if selected_sign_type == sign_type.IMAGE.value:
            if props.image_usage_option == 'IMPORT':
                # open filebrowser to select image
                # import image and create appropriate mesh for it.
                # create material with custom material node group
                # assign geometry node modifier with image node group
                bpy.ops.procedural_signs.append_images('INVOKE_DEFAULT')
            if props.image_usage_option == 'EXISTING':
                # create appropriate mesh for it.
                # create material with custom material node group
                # assign geometry node modifier with image node group
                bpy.ops.procedural_signs.append_images('INVOKE_DEFAULT',offset=False, is_ui_popup = True)
        if selected_sign_type == sign_type.OTHER.value:
            bpy.ops.procedural_signs.append_other('INVOKE_DEFAULT')
        return {"FINISHED"}

class Procedural_Signs_OT_Append_NEW_DRAW(Operator, Procedural_Signs_Append_Mixin):
    bl_idname = "procedural_signs.append_new_draw"
    bl_label = "Procedural Signs - Append New Draw"
    bl_description = "Description that shows in blender tooltips"
    bl_options = {"REGISTER","UNDO","INTERNAL"}

    @classmethod
    def poll(cls, context):
        return True
    
    def execute(self, context):
        props = context.scene.procedural_signs
        # create a empty curve object
        name = f'{props.sign}_sign'
        curve_data = bpy.data.curves.new(name,'CURVE')
        curve_obj = bpy.data.objects.new(name,curve_data)
        curve_obj.rotation_euler = (3.14/2,0,0)
        self.setup_objects_with_geometry_nodegroup(context, [curve_obj],"_Curve")

        # go to edit mode
        bpy.ops.object.select_all(action='DESELECT')
        curve_obj.select_set(True)
        context.view_layer.objects.active = curve_obj
        # go to top view
        bpy.ops.view3d.view_axis(type='FRONT')
        # start draw_curve operator
        bpy.ops.object.mode_set(mode='EDIT')
        bpy.ops.wm.tool_set_by_id(name="builtin.draw")

        return {"FINISHED"}

class Procedural_Signs_OT_Append_NEW_TEXT(Operator, Procedural_Signs_Append_Mixin):
    bl_idname = "procedural_signs.append_new_text"
    bl_label = "Procedural Signs - Append New Text"
    bl_description = "Description that shows in blender tooltips"
    bl_options = {"REGISTER","UNDO","INTERNAL"}

    @classmethod
    def poll(cls, context):
        return True
    
    def execute(self, context):
        props = context.scene.procedural_signs
        # create a empty curve object
        name = f'{props.sign}_sign'

        path_to_blend = self.get_path_to_sign_blendfile(context)
        with bpy.data.libraries.load(str(path_to_blend), link=False) as (data_from, data_to):
            data_to.curves = [curve for curve in data_from.curves if curve == f'{props.sign}_Text' ]
        
        curve_data = data_to.curves[0]
        curve_obj = bpy.data.objects.new(name,curve_data)
        curve_obj.rotation_euler = (3.14/2,0,0)
        self.setup_objects_with_geometry_nodegroup(context, [curve_obj],"_Text")

        # go to edit mode
        bpy.ops.object.select_all(action='DESELECT')
        curve_obj.select_set(True)
        context.view_layer.objects.active = curve_obj

        return {"FINISHED"}

class Procedural_Signs_OT_Append_SELECTED(Operator, Procedural_Signs_Append_Mixin):
    bl_idname = "procedural_signs.append_selected"
    bl_label = "Procedural Sign - Append Selected"
    bl_description = "Description that shows in blender tooltips"
    bl_options = {"REGISTER","UNDO","INTERNAL"}

    @classmethod
    def poll(cls, context):
        context.active_object is not None
        return True
    
    def check_modifier(self, obj:bpy.types.Object,type_:str, name:str) -> bool:
        for mod in obj.modifiers:
            if mod.type == type_ and mod.name == name:
                return True
        return False

    def execute(self, context):
        props = context.scene.procedural_signs

        curve_obj = context.active_object
        if self.check_modifier(curve_obj,'NODES',MODIFIER_NAME):
            self.report({'WARNING'},"Procedural Sign Moidifer already exists on this object")
            return {'CANCELLED'}

        self.setup_objects_with_geometry_nodegroup(context, [curve_obj],"_Curve",link_to_collection=False)
        return {"FINISHED"}

class Procedural_Signs_OT_Append_OTHER(Operator, Procedural_Signs_Append_Mixin):
    bl_idname = "procedural_signs.append_other"
    bl_label = "Procedural Sign - Append Other"
    bl_description = "Description that shows in blender tooltips"
    bl_options = {"REGISTER","UNDO","INTERNAL"}

    @classmethod
    def poll(cls, context):
        if context.mode != 'OBJECT':
            cls.poll_message_set('You should be in Object mode!')
            return False
        return True
    
    def execute(self, context):
        props = context.scene.procedural_signs
        path_to_blend = self.get_path_to_sign_blendfile(context)

        with bpy.data.libraries.load(str(path_to_blend), link=False) as (data_from, data_to):
            data_to.collections = [coll for coll in data_from.collections if coll in (OTHER_MAIN_COLLECTION,OTHER_ASSET_COLLECTION)]

        wrapper_coll = self.ensure_and_get_collection(f'Signs - {props.sign}')
        for collection in data_to.collections:
            if OTHER_MAIN_COLLECTION in collection.name:
                collection.name = props.sign
            wrapper_coll.children.link(collection)
        
        return {'FINISHED'}
    
class Procedural_Signs_OT_Add_Preset(Operator, Procedural_Signs_Append_Mixin):
    bl_idname = "procedural_signs.add_preser"
    bl_label = "Add Preset"
    bl_description = "Add Preset of Procedural Sign"
    bl_options = {"REGISTER","UNDO"}

    @classmethod
    def poll(cls, context):
        props = context.scene.procedural_signs
        if context.mode != 'OBJECT':
            cls.poll_message_set('You should be in Object mode!')
            return False
        return props.preset not in (None,'')
    
    def get_path_to_preset_blendfile(self,context):
        props = context.scene.procedural_signs
        asset_dir = context.preferences.addons[ROOT_PACKAGE].preferences.filepath
        signs_path = Path(asset_dir)/'Presets'/props.preset_category/props.preset
        path_to_blend = next(signs_path.glob("*.blend"))
        return path_to_blend
    
    def execute(self, context):
        props = context.scene.procedural_signs
        path_to_blend = self.get_path_to_preset_blendfile(context)

        with bpy.data.libraries.load(str(path_to_blend), link=False) as (data_from, data_to):
            data_to.collections = [coll for coll in data_from.collections if coll in (PRESET_MAIN_COLLECTION,PRESET_ASSET_COLLECTION)]

        wrapper_coll = self.ensure_and_get_collection(f'Preset - {props.preset}')
        for collection in data_to.collections:
            if PRESET_MAIN_COLLECTION in collection.name:
                collection.name = props.preset
            wrapper_coll.children.link(collection)
        
        for collection in data_to.collections:
            if PRESET_ASSET_COLLECTION in collection.name:
                exclude_collection(context,collection.name,exclude=True)
        return {'FINISHED'}

class Procedural_Signs_OT_Copy_Nodegroup(Operator, Procedural_Signs_Append_Mixin):
    bl_idname = "procedural_signs.copy_nodegroup"
    bl_label = "Single User Copy"
    bl_description = ("Make a Single User Copy of the geometry nodegroup.\
                    \nTo have a different settings from other objects,duplicating the nodegroup is required")
    bl_options = {"REGISTER","UNDO"}

    @classmethod
    def poll(cls, context):
        return hasattr(context,'current_modifier')
    
    def execute(self, context: Context):
        mod = context.current_modifier
        dup_mod = mod.node_group.copy()
        mod.node_group = dup_mod
        return {'FINISHED'}

class Procedural_Signs_OT_Open_Preferences(Operator, Procedural_Signs_Append_Mixin):
    bl_idname = "procedural_signs.open_preferences"
    bl_label = "Open Preferences..."
    bl_description = "Open Addon Preferences"
    bl_options = {"REGISTER","UNDO"}

    @classmethod
    def poll(cls, context):
        return True
    
    def execute(self, context: Context):
        context.preferences.active_section  = "ADDONS"
        context.window_manager.addon_search = "Procedural Signs"
        bpy.ops.screen.userpref_show("INVOKE_DEFAULT")
        bpy.ops.preferences.addon_expand(module=ROOT_PACKAGE)
        return {'FINISHED'}

classes = [
    Procedural_Signs_OT_append, 
    Procedural_Signs_OT_Append_NEW_DRAW, 
    Procedural_Signs_OT_Append_NEW_TEXT, 
    Procedural_Signs_OT_Append_OTHER,
    Procedural_Signs_OT_Append_SELECTED,
    Procedural_Signs_OT_Add_Preset,
    Procedural_Signs_OT_Copy_Nodegroup,
]
