import bpy
from bpy.types import Operator

from ..definitions import MODIFIER_NAME

def get_procedural_sign_modifier(obj):
    if obj is None:
        return -1, None
    for i, mod in enumerate(obj.modifiers):
        if mod.type == 'NODES' and MODIFIER_NAME in mod.name:
            return i, mod
    return -1, None

def get_decimate_after_sign_modifier(obj):
    if obj is None:
        return -1, None
    index, sign_modifier = get_procedural_sign_modifier(obj)
    if sign_modifier is None:
        return -1, None
    for i, mod in enumerate(obj.modifiers):
        if mod.type == 'DECIMATE' and i > index:
            return i, mod
    return -1, None

def convert_curve_to_mesh(context, curve_obj, index_of_decimate):
        
        # we only wants to apply modifiers till decimate [including it]
        # converting mesh operator takes account the visibility of modifier and if it is not visible then not included in the calculation.
        # after converting the mesh we reapply the remaining modifires
        excluded_modifiers = curve_obj.modifiers[index_of_decimate+1:]
        for modifier in excluded_modifiers:
            modifier.show_viewport = False
        
        context_ = context.copy()
        context_['object'] = context.active_object
        try:
            with bpy.context.temp_override(object=context.active_object):
                bpy.ops.object.convert(target='MESH', keep_original= True)
        except:
            raise 
        mesh_obj = context.active_object


        bpy.ops.object.select_all(action='DESELECT')
        context.view_layer.objects.active = curve_obj
        curve_obj.select_set(True)
        mesh_obj.select_set(True)

        for modifier in excluded_modifiers:
            modifier.show_viewport = True
            bpy.ops.object.modifier_copy_to_selected(modifier=modifier.name)

        context.view_layer.objects.active = mesh_obj
        mesh_obj.select_set(True)
        return mesh_obj

class Procedural_Signs_OT_Add_Modifier(Operator):
    bl_idname = "procedural_signs.add_modifier"
    bl_label = "Add Modifier"
    bl_description = ("Add Decimate Modifier after the Procedural Sign Geoemtry Node Modifier")
    bl_options = {'REGISTER', 'UNDO'}

    @classmethod
    def poll(cls, context):
        if context.active_object is None:
            return False
        index, sign_modifier = get_procedural_sign_modifier(context.active_object)
        if sign_modifier is None:
            return False
        return True

    def execute(self, context):
        index, sign_modifier = get_procedural_sign_modifier(context.active_object)
        for i, mod in enumerate(context.active_object.modifiers):
            if mod.type == 'DECIMATE' and i > index:
                self.report({'INFO'}, "Decimate modifier is already present")
                return {'CANCELLED'}
        decimate_modifier = context.active_object.modifiers.new("Decimate", 'DECIMATE')
        decimate_modifier.show_viewport = True
        # decimate_modifier.ratio = 0.5
        return {'FINISHED'}

class Procedural_Signs_OT_Apply_Modifier(Operator):
    bl_idname = "procedural_signs.apply_modifiers"
    bl_label = "Apply Modifiers"
    bl_description = ("Apply modifiers of the active object\n")
    bl_options = {'REGISTER', 'UNDO'}

    @classmethod
    def poll(cls, context):
        if context.active_object is None:
            return False
        index, modifier = get_decimate_after_sign_modifier(context.active_object)
        if modifier is None:
            return False
        return True

    def execute(self, context):
        message_a= ""
        # collect names for objects failed to apply modifiers

        curve_obj = context.active_object
        index, modifier = get_decimate_after_sign_modifier(context.active_object)

        # copying context for the operator's override
        try:
            mesh_obj = convert_curve_to_mesh(context, curve_obj, index)
        except:
            self.report({'ERROR'}, "Error while applying the modifiers")
            return {'FINISHED'}
        bpy.data.objects.remove(curve_obj)

        message_a = "Applying modifiers on Active Object"
        self.report({"INFO"},(message_a))

        return {'FINISHED'}

classes = [
    Procedural_Signs_OT_Add_Modifier,
    Procedural_Signs_OT_Apply_Modifier
]