from collections import defaultdict
import bpy
from bpy.types import UILayout, NodesModifier
from dataclasses import dataclass

_ui_item_state = defaultdict(lambda: False)

@dataclass
class Signs_Pointer_Prop:
    type_:str
    search_id:str
    icon:str

def nodetree_inputs(tree):
    import bpy

    if bpy.app.version < (4, 0, 0):
        return tree.inputs
    else:

        def inputs():
            for item in tree.interface.items_tree:
                if item.item_type == "SOCKET" and item.in_out == "INPUT":
                    yield item

        iterator = inputs()
        return iterator


def get_socket_type_value(key: str):
    import bpy

    key = key.upper()
    lut = {
        "STRING": ["STRING", "NodeSocketString"],
        "BOOLEAN": ["BOOLEAN", "NodeSocketBool"],
        "MATERIAL": ["MATERIAL", "NodeSocketMaterial"],
        "VECTOR": ["VECTOR", "NodeSocketVector"],
        "INT": ["INT", "NodeSocketInt"],
        "GEOMETRY": ["GEOMETRY", "NodeSocketGeometry"],
        "COLLECTION": ["COLLECTION", "NodeSocketCollection"],
        "TEXTURE": ["TEXTURE", "NodeSocketTexture"],
        "VALUE": ["VALUE", "NodeSocketFloat"],
        "RGBA": ["RGBA", "NodeSocketColor"],
        "OBJECT": ["OBJECT", "NodeSocketObject"],
        "IMAGE": ["IMAGE", "NodeSocketImage"],
        "ROTATION": ["N/I", "NodeSocketRotation"],
    }
    if bpy.app.version < (4, 0, 0):
        return lut[key][0]
    else:
        return lut[key][1]


def get_socket_type(socket):
    import bpy

    if bpy.app.version < (4, 0, 0):
        return socket.type
    else:
        return socket.socket_type



pointer_data:dict = {
    get_socket_type_value("OBJECT"): Signs_Pointer_Prop("OBJECT", "objects", "OBJECT_DATA"),
    get_socket_type_value("COLLECTION"): Signs_Pointer_Prop("COLLECTION", "collections", "OUTLINER_COLLECTION"),
    get_socket_type_value("MATERIAL"): Signs_Pointer_Prop("MATERIAL", "materials", "MATERIAL"),
    get_socket_type_value("TEXTURE"): Signs_Pointer_Prop("TEXTURE", "textures", "TEXTURE"),
    get_socket_type_value("IMAGE"): Signs_Pointer_Prop("IMAGE", "images", "IMAGE"),
}

curve_type = {
    'CURVE_VEC' : 'VECTOR',
    'CURVE_RGB' : 'COLOR',
    'CURVE_FLOAT' : 'NONE',
} 

def get_gn_mod(obj: bpy.types.Object, mod_name):
    mod = None
    for m in obj.modifiers:
        if m.type == "NODES" and m.name == mod_name:
            mod = m
            break
    return mod


def draw_gn_mod_props(layout:UILayout,root:UILayout,mod:NodesModifier,input):
    data, prop, use_attr, prop_attr = (
        mod, 
        f'["{input.identifier}"]', 
        f'["{input.identifier}_use_attribute"]', 
        f'["{input.identifier}_attribute_name"]'
    )
  
    if input.name == '' and get_socket_type(input) == get_socket_type_value('STRING'):
        box = root.box()
        box.operator("draw_geonode_ui.property_toggle",
                        text=input.default_value,
                        icon="TRIA_DOWN" if _ui_item_state[input.identifier] else "TRIA_RIGHT",
                        emboss=False,
                    ).api = input.identifier
        if _ui_item_state[input.identifier]:
            col = box.column(align=True)
            return col
        else:
            return None

    if layout is None:
        return None
    if input.name.startswith('PS.'):
        metadata,name = input.name.rsplit('_',1)
        actions = metadata.split('.')[1:]
        if len(actions)== 1 and actions[0] == 'CURVE':
            node = mod.node_group.nodes.get(name)
            if node is None:
                box= layout.box()
                box.label(text=f"Can not find the node with the name: {node.name}",icon='ERROR')
            box = layout.box()
            row = box.split(factor=0.8 if mod.node_group.users > 1 else 1)
            row.label(text='',icon='BLANK1')
            if mod.node_group.users > 1:
                row.context_pointer_set('current_modifier', mod)
                row.operator('procedural_signs.copy_nodegroup', text=str(mod.node_group.users), icon='NODETREE')
            subbox = box.box()
            subbox.template_curve_mapping(node,'mapping',type=curve_type[node.type])
        if len(actions)== 1 and actions[0] == 'FONT':
            node = mod.node_group.nodes.get(name)
            if node is None:
                box= layout.box()
                box.label(text=f"Can not find the node with the name: {node.name}",icon='ERROR')
            box = layout.box()
            row = box.split(factor=0.8 if mod.node_group.users > 1 else 1)
            row.template_ID(node,'font',open='font.open')
            if mod.node_group.users > 1:
                row.context_pointer_set('current_modifier', mod)
                row.operator('procedural_signs.copy_nodegroup', text=str(mod.node_group.users), icon='NODETREE')
        if len(actions)>1:
            node = mod.node_group.nodes.get(name)
            if node is None:
                box= layout.box()
                box.label(text=f"Can not find the node with the name: {node.name}",icon='ERROR')
            box = box.box()
            box.label(text=f"Naming format is not supported. It seems name has multiple actions.",icon='ERROR')
        return layout

    box = layout.box()
    box.use_property_split = True
    box.use_property_decorate = True

    row = box.split(align=True, factor=0.4)
    row.alignment='RIGHT'
    row.label(text=input.name)
    
    subrow = row.row(align=True)
    subrow.use_property_split=True
    subrow.use_property_decorate=True

    if 'DIAMOND' in mod.node_group.nodes['Group Input'].outputs[input.name].display_shape: # Possible shape : Diamond, Diamond_dot
        op_props = subrow.operator('object.geometry_nodes_input_attribute_toggle',text='', icon='SPREADSHEET')
        if bpy.app.version < (4, 0, 0):
            op_props.prop_path = use_attr
        else:          
            op_props.input_name = input.identifier
        op_props.modifier_name = mod.name

    if get_socket_type(input) in pointer_data.keys():
        subrow.prop_search(
            mod,
            prop,
            bpy.data,
            pointer_data[get_socket_type(input)].search_id,
            text='',
            icon=pointer_data[get_socket_type(input)].icon)
    else:
        final_prop = prop_attr if getattr(mod,use_attr,0)!=0 else prop
        subrow.prop(data,final_prop, text='',)
    return layout

class SIGNS_DRAW_GEONODE_UI_OT_property_toggle(bpy.types.Operator):
    """useful for custom gui, as blender will register simple boolean toggle to undo history"""

    bl_idname      = "draw_geonode_ui.property_toggle"
    bl_label       = ""
    bl_description = ""
    #DO NOT REGISTER TO UNDO HERE!!!

    api : bpy.props.StringProperty()

    def execute(self, context):
        global _ui_item_state
        if self.api not in _ui_item_state:
            _ui_item_state[self.api] = True
        _ui_item_state[self.api] = not _ui_item_state[self.api]
        return {'FINISHED'}