from dataclasses import dataclass
from enum import Enum
from typing import Optional
import bpy
from bpy.types import PropertyGroup, Scene
from bpy.props import EnumProperty, PointerProperty, BoolProperty
from pathlib import Path
from .t3dn_bip import previews
import json



_item_map = dict()
_sign_metadata_map = dict()

def _make_item(enum_name, id_, name, descr, preview_id, uid):
    lookup = f"{str(id_)}\0{str(name)}\0{str(descr)}\0{str(preview_id)}\0{str(uid)}"
    if enum_name not in _item_map:
        _item_map[enum_name] = {}
    if lookup not in _item_map[enum_name]:
        _item_map[enum_name][lookup] = (id_, name, descr, preview_id, uid)
    return _item_map[enum_name][lookup]

def read_json(path):
    '''read json file and returns python equivalent object'''
    with open(path,'r') as f:
        data = json.load(f)
    return data

class sign_type(Enum):
    IMAGE =  "IMAGE"
    OBJECT = "OBJECT"
    OTHER = "OTHER"

@dataclass
class Metadata:
    sign_type:sign_type
    min_required_version:list
    material_nodegroup_name: Optional[str] = ''



class Procedural_Signs_Properties(PropertyGroup):
    import_options = (
        ('TEXT','Text','Import Text gn ng'),
        ('SVG','SVG',''),
        ('DRAW','Draw','')
    )
    object_usage_options=(
        ('NEW','Create New',''),
        ('SELECTED','Use Selected',''),
    )
    image_usage_options=(
        ('IMPORT','Import Image',''),
        ('EXISTING','Use Existing',''),
    )

    @staticmethod
    def get_thumbnail(dir:Path, default=0):
        global thumbnails
        thumbnail_path = next(dir.glob("*.png"), None)
        thumbnail = default
        if thumbnail_path:
            thumbnail = thumbnails.load_safe(dir.name.lower(),str(thumbnail_path),'IMAGE').icon_id
        return thumbnail

    @staticmethod
    def read_metadata(dir:Path,filename='metadata'):
        key = f'Signs_{dir.name}'
        if key in _sign_metadata_map:
            return _sign_metadata_map[key]
        
        file = f'{filename}.json'
        file = next(dir.glob(file),None)
        if file:
            _sign_metadata_map[key] = Metadata(**read_json(file))

    def get_signs(self, context):
        enum_items = []
        signs_dir = context.preferences.addons[__package__].preferences.filepath
        if not signs_dir:
            return enum_items

        path = Path(signs_dir) / 'Signs'
        for i, dir in enumerate(path.iterdir()):
            if not dir.is_dir():
                continue
            thumbnail = self.get_thumbnail(dir, default='SYSTEM')
            enum_items.append(_make_item('Signs',dir.name, dir.name, '', thumbnail, i))
            self.read_metadata(dir)
        return enum_items

    sign:EnumProperty(items=get_signs)
    import_option:EnumProperty(items=import_options,name='Import Option',description="Import Options")
    object_usage_option:EnumProperty(items=object_usage_options,name='Object Usage Option')
    image_usage_option:EnumProperty(items=image_usage_options,name='Image Usage Option')
    selected_image:PointerProperty(type=bpy.types.Image)


    def get_presets(self, context):
        enum_items = []
        asset_dir = context.preferences.addons[__package__].preferences.filepath
        if not asset_dir:
            return enum_items

        props = context.scene.procedural_signs
        path = Path(asset_dir) / 'Presets' / props.preset_category
        for i, dir in enumerate(path.iterdir()):
            if not dir.is_dir():
                continue
            thumbnail = self.get_thumbnail(dir, default='SYSTEM')
            enum_items.append(_make_item('Presets',dir.name, dir.name, '', thumbnail, i))
        return enum_items

    def get_preset_categories(self, context):
        enum_items = []
        asset_dir = context.preferences.addons[__package__].preferences.filepath
        if not asset_dir:
            return enum_items

        path = Path(asset_dir) / 'Presets'
        for i, dir in enumerate(path.iterdir()):
            if not dir.is_dir():
                continue
            thumbnail = self.get_thumbnail(dir, default='SYSTEM')
            enum_items.append(_make_item('preset_category',dir.name, dir.name, '', thumbnail, i))
        return enum_items

    preset:EnumProperty(items=get_presets, name='Preset', description='Sign Preset to import')
    preset_category:EnumProperty(items=get_preset_categories, name='Category', description='Category of Sign Presets')

classes = [Procedural_Signs_Properties]

def register():
    global thumbnails
    thumbnails = previews.new(max_size=(128, 128))
    for cls in classes:
        bpy.utils.register_class(cls)
    Scene.procedural_signs = PointerProperty(type=Procedural_Signs_Properties)

def unregister():
    global thumbnails
    if thumbnails:
        del thumbnails
    del Scene.procedural_signs
    for cls in reversed(classes):
        bpy.utils.unregister_class(cls)