from dataclasses import dataclass
import bpy
from bpy.types import Panel, UILayout, NodesModifier
import webbrowser
import os
import json

from .definitions import MODIFIER_NAME
from .properties import _sign_metadata_map, sign_type
from .Operators.optimize_modifiers import get_decimate_after_sign_modifier
from .draw_geonode_ui import (
    nodetree_inputs,
    draw_gn_mod_props,
    get_gn_mod,
    SIGNS_DRAW_GEONODE_UI_OT_property_toggle
)
class Procedural_Signs_PT_Browser(Panel):

    bl_idname = "PROCEDURAL_SIGNS_PT_Browser"
    bl_label = "Procedural Signs"
    bl_space_type = 'VIEW_3D'
    bl_region_type = 'UI'
    bl_category = 'Procedural Signs'

    def draw(self, context):
        layout = self.layout
        props = context.scene.procedural_signs
        scale_y = 1.4
        icon_scale = 5
        scale_popup = 6

        signs_dir = bpy.context.preferences.addons[__package__].preferences.filepath
        if not signs_dir:
            box=layout.box()
            box.label(text="Select Asset Folder in Preferences", icon='ERROR')
            box.operator('procedural_signs.open_preferences')
            return
        
        box = layout.box()
        row = box.row()
        row.template_icon_view(props,"sign", show_labels=True)
        row = box.row()
        row.alignment = 'CENTER'
        row.label(text=props.sign.capitalize())

        if _sign_metadata_map[f'Signs_{props.sign}'].sign_type == sign_type.IMAGE.value:
            row = box.row()
            row.prop(props,"image_usage_option",expand=True)
            row = box.row()
            row.enabled = props.image_usage_option == 'EXISTING'
            row.prop(props,"selected_image", text='')
            row = box.row()
            row.operator('procedural_signs.append')
        if _sign_metadata_map[f'Signs_{props.sign}'].sign_type == sign_type.OBJECT.value:
            row = box.row()
            row.prop(props,"object_usage_option",expand=True)
            row = box.row()
            row.enabled = props.object_usage_option == 'NEW'
            row.prop(props,"import_option",expand=True)
            row = box.row()
            row.operator('procedural_signs.append')
        if _sign_metadata_map[f'Signs_{props.sign}'].sign_type == sign_type.OTHER.value:
            row = box.row()
            # row.label(text="Not Implemented")
            row.operator('procedural_signs.append')

class Procedural_Signs_PT_GeoNode(Panel):

    bl_idname = "PROCEDURAL_SIGNS_PT_GeometryNodes"
    bl_label = "Signs Settings"
    bl_space_type = 'VIEW_3D'
    bl_region_type = 'UI'
    bl_category = 'Procedural Signs'
    bl_options = {'DEFAULT_CLOSED'}

    def draw(self, context):
        layout: UILayout = self.layout
        if not context.active_object:
            box = layout.box()
            box.label(text="No Active Object", icon='INFO')
            return
        
        mod = get_gn_mod(context.active_object, MODIFIER_NAME)
        if not mod:
            box = layout.box()
            box.label(text="No Procedural Sign System Found.", icon='INFO')
            return

        col = layout.column(align=True)
        _layout = col
        for input in nodetree_inputs(mod.node_group):
            if input.identifier == "Input_0": #skip the first inout which is always Geoemetry Input
                continue
            _layout = draw_gn_mod_props(_layout,col,mod,input)

class Procedural_Signs_PT_Optimize(Panel):

    bl_idname = "PROCEDURAL_SIGNS_PT_Optimize"
    bl_label = "Optimize"
    bl_space_type = 'VIEW_3D'
    bl_region_type = 'UI'
    bl_category = 'Procedural Signs'

    decimate_type_prop_map = {
            "COLLAPSE" : "ratio",
            "UNSUBDIV" : "iterations",
            "DISSOLVE" : "angle_limit"
    }
    def draw(self, context):
        layout: UILayout = self.layout
        
        row = layout.row()
        row.operator('procedural_signs.add_modifier', text='Decimate')
        index,modifier = get_decimate_after_sign_modifier(context.active_object)
        if modifier is not None:
            box = layout.box()
            box.prop(modifier, "decimate_type")
            modifier_prop = self.decimate_type_prop_map[modifier.decimate_type]
            box.prop(modifier, modifier_prop)
            box.label(text=f'Face Count: {modifier.face_count}')
            row = layout.row()
            row.operator('procedural_signs.apply_modifiers')

class Procedural_Signs_PT_Presets(Panel):

    bl_idname = "PROCEDURAL_SIGNS_PT_Presets"
    bl_label = "Presets"
    bl_space_type = 'VIEW_3D'
    bl_region_type = 'UI'
    bl_category = 'Procedural Signs'

    def draw(self, context):
        layout = self.layout
        props = context.scene.procedural_signs
        scale_y = 1.4
        icon_scale = 5
        scale_popup = 6

        signs_dir = context.preferences.addons[__package__].preferences.filepath
        if not signs_dir:
            box=layout.box()
            box.label(text="Select Asset Folder in Preferences", icon='ERROR')
            box.operator('procedural_signs.open_preferences')
            return
        
        row = layout.row()
        row.prop(props,"preset_category")
        row = layout.row()
        row.template_icon_view(props,"preset", show_labels=True)
        row = layout.row()
        row.alignment = 'CENTER'
        row.label(text=props.preset)
        row = layout.row()
        row.operator('procedural_signs.add_preser')



class Procedural_Signs_PT_Help(Panel):
    bl_idname = "PROCEDURAL_SIGNS_PT_Help"
    bl_label = "Help"
    bl_space_type = 'VIEW_3D'
    bl_region_type = 'UI'
    bl_category = 'Procedural Signs'
    bl_options = {'DEFAULT_CLOSED'}

    @classmethod
    def poll(cls, context):
        return True

    def draw(self, context):
        dns = bpy.app.driver_namespace
        links = dns.get("signs_links")
        if not links:
            get_links()
        layout = self.layout
        
        if links:
            for link in links:
                label = link["name"]
                url = link["link"]
                icon = link["icon"]
                row = layout.row()
                row.operator("procedural_signs.url_open", text=label, icon=icon).url = url



class Procedural_Signs_OT_OpenURL(bpy.types.Operator):
    bl_idname = "procedural_signs.url_open"
    bl_label = "Open URL"

    url: bpy.props.StringProperty()

    def execute(self, context):
        webbrowser.open(self.url)
        return {'FINISHED'}



def get_links():
    dns = bpy.app.driver_namespace
    dns["signs_links"] = {}
    current_folder = os.path.dirname(os.path.realpath(__file__))
    links_json = os.path.join(current_folder, "links.json")
    if os.path.exists(links_json):
        with open(links_json, "r") as f:
            dns["signs_links"] = json.load(f)

classes = [
    Procedural_Signs_PT_Browser,
    Procedural_Signs_PT_GeoNode,
    Procedural_Signs_PT_Optimize,
    Procedural_Signs_PT_Presets,
    SIGNS_DRAW_GEONODE_UI_OT_property_toggle,
    Procedural_Signs_PT_Help,
    Procedural_Signs_OT_OpenURL,
    ]

def register():
    for c in classes:
        bpy.utils.register_class(c)

def unregister():
    for c in reversed(classes):
        bpy.utils.unregister_class(c)
